#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file WiFiNeighborNodeViewer.py
 @brief This RTC displays neighbor Wi-Fi nodes information.
 @date $Date$

 @author 株式会社 東日本計算センター

 プロジェクト方針決定後に記載

"""
import sys
import copy
from logging import getLogger, StreamHandler
from logging import INFO as LOG_LV

# Import RTM module
import RTC
import OpenRTM_aist
import WiFiNodeIF
from tkviewer import TkWiFiNeighborNodeViewer, OBSERVER_NODE, NEIGHBOR_NODE, STATUS, ALL

sys.path.append(".")

# Setup logger
logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(LOG_LV)
handler.setLevel(LOG_LV)
logger.addHandler(handler)

# Setup tk_viewer
tk_viewer = TkWiFiNeighborNodeViewer()

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
wifineighbornodeviewer_spec = ["implementation_id", "WiFiNeighborNodeViewer",
                               "type_name",         "WiFiNeighborNodeViewer",
                               "description",
                               "This RTC displays neighbor Wi-Fi nodes information.",
                               "version",           "0.1",
                               "vendor",            "EAC",
                               "category",          "Network",
                               "activity_type",     "STATIC",
                               "max_instance",      "1",
                               "language",          "Python",
                               "lang_type",         "SCRIPT",
                               ""]
# </rtc-template>


##
# @class WiFiNeighborNodeViewer
# @brief This RTC displays neighbor Wi-Fi nodes information.
#
# 別紙"機能仕様書_WiFiNeighborNodeViewerRTC"参照
#
# 別紙"機能仕様書_WiFiNeighborNodeViewerRTC"参照
#
# 別紙"機能仕様書_WiFiNeighborNodeViewerRTC"参照
#
# 別紙"機能仕様書_WiFiNeighborNodeViewerRTC"参照
#
#
class WiFiNeighborNodeViewer(OpenRTM_aist.DataFlowComponentBase):
    ##
    # @brief constructor
    # @param manager Maneger Object
    #

    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_WiFiNeighborNodeListIn = WiFiNodeIF.TimedNeighborNodeList(
            RTC.Time(0, 0), 0, [])
        """
        """
        self._WiFiNeighborNodeListInIn = OpenRTM_aist.InPort("WiFiNeighborNodeListIn",
                                                             self._d_WiFiNeighborNodeListIn)

        self._init_variable()

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">

        # </rtc-template>

    ##
    # RTC初期化
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    #
    def onInitialize(self):
        # Bind variables and configuration variable

        # Set InPort buffers
        self.addInPort(
            "WiFiNeighborNodeListIn", self._WiFiNeighborNodeListInIn)

        # Set OutPort buffers

        # Set service provider to Ports

        # Set service consumers to Ports

        # Set CORBA Service Ports

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The finalize action (on ALIVE->END transition)
    #    # formaer rtc_exiting_entry()
    #    #
    #    # @return RTC::ReturnCode_t
    #
    #    #
    # def onFinalize(self):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK

        ##
        # データリスナー登録
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):
        self._listener = DataListener(self._cb_receive_neighbor_node_data)
        listener_type = OpenRTM_aist.ConnectorDataListenerType.ON_RECEIVED

        self._WiFiNeighborNodeListInIn.addConnectorDataListener(
            listener_type, self._listener)

        return RTC.RTC_OK

        ##
        # データリスナー登録解除
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):
        listener_type = OpenRTM_aist.ConnectorDataListenerType.ON_RECEIVED

        self._WiFiNeighborNodeListInIn.removeConnectorDataListener(
            listener_type, self._listener)

        self._init_variable()
        tk_viewer.init_viewer()

        return RTC.RTC_OK

        ##
        # Wi-Fi近隣ノード情報受信チェック周期処理
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):
        self._check_missing_observer_node()

        return RTC.RTC_OK

        ##
        # エラーログ出力
        #
        # The aborting action when main logic error occurred.
        # former rtc_aborting_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onAborting(self, ec_id):
        logger.error("ERROR: detected an Aborting state")

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The error action in ERROR state
    #    # former rtc_error_do()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onError(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The reset action that is invoked resetting
    #    # This is same but different the former rtc_init_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onReset(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    ##
    # def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK

    ##
    # @brief データ受信用変数初期化。
    #
    # データ受信用変数初期化。
    def _init_variable(self):
        self._neighbor_node_list = []
        self._neighbor_node_list_bk = []
        self._listener = None

    ##
    # @brief ObserverのIDが、リストに載っていれば受信データを返す。
    #
    # ロボットからの報告を受信する。
    #
    # @return ロボットからの受信データ
    def get_neighbor_node_data(self, observer_node_id):
        ret = None
        for data in self._neighbor_node_list:
            if observer_node_id == data.id:
                ret = data
                break

        if not ret:
            logger.error(
                "invalid requested observer_node_id: " + observer_node_id)

        return ret

    ##
    # @brief ロボットからの報告が無ければ、エラーを表示する。
    #
    # ロボットからの報告が無ければ、エラーを表示する。
    def _check_missing_observer_node(self):
        for i, (r1, r2) in enumerate(zip(self._neighbor_node_list,
                                         self._neighbor_node_list_bk)):
            if r1 == r2:
                if r1.data:
                    del r1.data[:]
                    tk_viewer.draw_gui(ALL, r1.id, missing=True)
                    logger.info("missing observer_node(%d): %s", i, r1.id)

        self._neighbor_node_list_bk = copy.deepcopy(self._neighbor_node_list)

    ##
    # @brief コールバックメソッド。データポートがデータを受け取ったら実行される。
    #
    # 以前のデータと比較して
    #
    # @param receive_neighbor_node_list
    def _cb_receive_neighbor_node_data(self, receive_neighbor_node_list):
        is_new = True
        for i, neighbor_node in enumerate(self._neighbor_node_list):
            if neighbor_node.id == receive_neighbor_node_list.id:
                self._neighbor_node_list[i] = copy.deepcopy(
                    receive_neighbor_node_list)
                is_new = False

                if neighbor_node.data == receive_neighbor_node_list.data:
                    # no neighbor_node change
                    tk_viewer.draw_gui(STATUS, neighbor_node.id)
                elif neighbor_node.data:
                    tk_viewer.draw_gui(
                        NEIGHBOR_NODE | STATUS, neighbor_node.id)
                    logger.info("changed the neighbor_node data of observer_node(%d): %s",
                                i, neighbor_node.id)
                    logger.info(receive_neighbor_node_list)
                else:
                    tk_viewer.draw_gui(ALL, neighbor_node.id, missing=False)
                    logger.info(
                        "found the missing observer_node(%d): %s", i, neighbor_node.id)
                    logger.info(receive_neighbor_node_list)

                break

        if is_new:
            self._neighbor_node_list.append(receive_neighbor_node_list)
            logger.info("new observer_node(%d): %s",
                        len(self._neighbor_node_list), receive_neighbor_node_list.id)
            logger.info(receive_neighbor_node_list)

            tk_viewer.draw_gui(OBSERVER_NODE, receive_neighbor_node_list.id)


##
# @class
# @brief
class DataListener(OpenRTM_aist.ConnectorDataListenerT):

    def __init__(self, func):
        self._plugin_func = func
        logger.info("added DataListener")

    def __del__(self):
        logger.info("removed DataListener")

    def __call__(self, info, cdrdata):
        neighbor_node = WiFiNodeIF.TimedNeighborNodeList(RTC.Time(0, 0), 0, [])
        neighbor_node_list = OpenRTM_aist.ConnectorDataListenerT.__call__(self, info,
                                                                          cdrdata,
                                                                          neighbor_node)

        self._plugin_func(neighbor_node_list)


def WiFiNeighborNodeViewerInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=wifineighbornodeviewer_spec)
    manager.registerFactory(profile,
                            WiFiNeighborNodeViewer,
                            OpenRTM_aist.Delete)


def MyModuleInit(manager):
    WiFiNeighborNodeViewerInit(manager)

    # Create a component
    comp = manager.createComponent("WiFiNeighborNodeViewer")

    tk_viewer.set_data_src_func(comp.get_neighbor_node_data)


def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager(True)
    tk_viewer.mainloop()
    mgr.shutdown()

if __name__ == "__main__":
    main()
